import os
from functools import partial, wraps
from pathlib import Path
from typing import Optional

from giza_actions.utils import get_workspace_uri  # noqa: E402

os.environ["PREFECT_API_URL"] = f"{get_workspace_uri()}/api"
os.environ["PREFECT_UI_URL"] = get_workspace_uri()

from prefect import Flow  # noqa: E402
from prefect import flow as _flow  # noqa: E402
from prefect.client.schemas.schedules import construct_schedule  # noqa: E402
from prefect.settings import PREFECT_API_URL  # noqa: E402
from prefect.settings import (  # noqa: E402
    PREFECT_LOGGING_SETTINGS_PATH,
    PREFECT_UI_URL,
    update_current_profile,
)
from prefect.utilities.asyncutils import sync_compatible  # noqa: E402
from rich.console import Console  # noqa: E402
from rich.panel import Panel  # noqa: E402

from giza_actions.__init__ import __module_path__  # noqa: E402


class Action:
    """
    A class to represent an Action.

    Attributes:
        name (str): The name of the action.
        _flow (Flow): The Prefect flow that represents the action's entrypoint.

    Methods:
        _set_settings: Updates the current profile with API and logging settings.
        _update_api_url: Updates the API URL in the current profile.
        get_flow: Returns the Prefect flow.
        serve: Serves the action, making it ready to poll for scheduled runs.
    """

    def __init__(self, entrypoint: Flow, name: str):
        """
        Constructs all the necessary attributes for the Action object.

        Args:
            entrypoint (Flow): The Prefect flow that represents the action's entrypoint.
            name (str): The name of the action.
        """
        self.name = name
        self._flow = entrypoint
        self._set_settings()

    def _set_settings(self):
        """
        Updates the current profile with the workspace API URL and the path to the logging configuration.
        """
        update_current_profile(settings={PREFECT_API_URL: f"{get_workspace_uri()}/api"})
        update_current_profile(
            settings={PREFECT_LOGGING_SETTINGS_PATH: f"{__module_path__}/logging.yaml"}
        )

    def _update_api_url(self, api_url: str):
        """
        Updates the API URL in the current profile.

        Args:
            api_url (str): The new API URL to set in the profile.
        """
        update_current_profile(settings={PREFECT_API_URL: api_url})

    def get_flow(self):
        """
        Returns the Prefect flow associated with the action.

        Returns:
            Flow: The Prefect flow of the action.
        """
        return self._flow

    @sync_compatible
    async def serve(
        self,
        name: str,
        cron: Optional[str] = None,
        interval: Optional[str] = None,
        parameters: Optional[dict] = None,
        print_starting_message: bool = True,
    ):
        """
        Serves the action, making it ready to poll for scheduled runs.

        Args:
            name (str): The name to assign to the runner. If a file path is provided, it uses the file name without the extension.
            print_starting_message (bool, optional): Whether to print a starting message. Defaults to True.
            interval: An interval on which to schedule runs. Accepts either a number
            or a timedelta object. If a number is given, it will be interpreted as seconds.
            cron: A cron schedule for runs.
        """

        workspace_url = get_workspace_uri()
        if workspace_url == "":
            raise ValueError(
                "Workspace URL cannot be empty. Please create a workspace using `giza workspace create` and wait for the workspace to have status COMPLETED."
            )

        from prefect.runner import Runner

        # Handling for my_flow.serve(__file__)
        # Will set name to name of file where my_flow.serve() without the extension
        # Non filepath strings will pass through unchanged
        name = Path(name).stem

        schedule = None

        if interval or cron:
            schedule = construct_schedule(
                interval=interval,
                cron=cron,
            )

        runner = Runner(name=name, pause_on_shutdown=False)
        deployment_id = await runner.add_flow(
            self._flow,
            name=name,
            schedule=schedule,
            parameters=parameters,
        )
        if print_starting_message:
            help_message = (
                f"[green]Your action {self.name!r} is being served and polling for"
                " scheduled runs!\n[/]"
            )
            if PREFECT_UI_URL:
                help_message += (
                    "\nYou can run your action via the Actions UI:"
                    f" [blue]{PREFECT_UI_URL.value()}/deployments/deployment/{deployment_id}[/]\n"
                )

            console = Console()
            console.print(Panel(help_message))
        await runner.start(webserver=False)


def action(func=None, *task_init_args, **task_init_kwargs):
    """
    Decorator to convert a function into a Prefect flow.

    Args:
        func (Callable, optional): The function to convert into a flow. If None, returns a partial function.
        **task_init_kwargs: Arbitrary keyword arguments passed to the flow initialization.

    Returns:
        Flow: The Prefect flow created from the function.
    """
    if func is None:
        return partial(action, *task_init_args, **task_init_kwargs)

    @wraps(func)
    def safe_func(*args, **kwargs):
        """
        A wrapper function that calls the original function with its arguments.

        Args:
            **kwargs: Arbitrary keyword arguments passed to the original function.

        Returns:
            The return value of the original function.
        """
        return func(*args, **kwargs)

    safe_func.__name__ = func.__name__
    return _flow(safe_func, *task_init_args, **task_init_kwargs)



import logging

import requests
from giza import API_HOST
from giza.client import EndpointsClient, WorkspaceClient

logger = logging.getLogger(__name__)


def get_workspace_uri():
    """
    Retrieves the URI of the current workspace.

    This function creates a WorkspaceClient instance using the API_HOST and
    calls its get method to retrieve the current workspace. It then returns
    the URL of the workspace.

    Returns:
        str: The URL of the current workspace.
    """
    client = WorkspaceClient(API_HOST)
    try:
        workspace = client.get()
    except requests.exceptions.RequestException:
        logger.error("Failed to retrieve workspace")
        logger.error(
            "Please check that you have create a workspaces using the Giza CLI"
        )
        raise
    return workspace.url


def get_endpoint_uri(model_id: int, version_id: int):
    """
    Get the deployment URI associated with a specific model and version.

    Args:
        model_id (int): The ID of the model.
        version_id (int): The ID of the version.

    This function initializes a DeploymentsClient instance using the API_HOST and
    retrieves the deployment URI using its list method. The resulting URL of the
    deployment is returned.

    Returns:
        str: The URI of the deployment.
    """
    client = EndpointsClient(API_HOST)
    deployments_list = client.list(
        params={"model_id": model_id, "version_id": version_id, "is_active": True}
    )

    if len(deployments_list.root) == 1:
        return deployments_list.root[0].uri
    else:
        return None



import logging
from pathlib import Path
from typing import Dict, Optional

import numpy as np
import onnxruntime as ort
import onnx
import requests
from giza import API_HOST
from giza.client import ApiClient, ModelsClient, VersionsClient
from giza.utils.enums import Framework, VersionStatus
from osiris.app import (
    create_tensor_from_array,
    deserialize,
    load_data,
    serialize,
    serializer,
)

from giza_actions.utils import get_endpoint_uri


class GizaModel:
    """
    A class to manage the lifecycle and predictions of models using both local ONNX runtime sessions and
    remote deployments via the Giza SDK.

    Attributes:
        session (ort.InferenceSession | None): An ONNX runtime inference session for local model predictions.
        model_client (ModelsClient): Client to interact with the models endpoint of the Giza API.
        version_client (VersionsClient): Client to interact with the versions endpoint of the Giza API.
        api_client (ApiClient): General client for interacting with the Giza API.
        uri (str): The URI for making prediction requests to a deployed model.

    Args:
        model_path (Optional[str]): The file path to a local ONNX model. Defaults to None.
        id (Optional[int]): The unique identifier of the model in the Giza platform. Defaults to None.
        version (Optional[int]): The version number of the model in the Giza platform. Defaults to None.
        output_path (Optional[str]): The file path where the downloaded model should be saved. Defaults to None.

    Raises:
        ValueError: If the necessary combination of parameters is not provided.
    """

    def __init__(
        self,
        model_path: Optional[str] = None,
        id: Optional[int] = None,
        version: Optional[int] = None,
        output_path: Optional[str] = None,
    ):
        if model_path is None and id is None and version is None:
            raise ValueError(
                "Either model_path or id and version must be provided.")

        if model_path is None and (id is None or version is None):
            raise ValueError("Both id and version must be provided.")

        if model_path and (id or version):
            raise ValueError(
                "Either model_path or id and version must be provided.")

        if model_path and id and version:
            raise ValueError(
                "Only one of model_path or id and version should be provided."
            )

        if model_path:
            self.session = ort.InferenceSession(model_path)
        elif id and version:
            self.model_client = ModelsClient(API_HOST)
            self.version_client = VersionsClient(API_HOST)
            self.api_client = ApiClient(API_HOST)
            self._get_credentials()
            self.model = self._get_model(id)
            self.version = self._get_version(version)
            self.session = self._set_session()
            self.framework = self.version.framework
            self.uri = self._retrieve_uri(version)
            if output_path:
                self._download_model(output_path)

    def _retrieve_uri(self, version_id: int):
        """
        Retrieves the URI for making prediction requests to a deployed model.

        Args:
            version_id (int): The version number of the model.

        Returns:
            The URI for making prediction requests to the deployed model.
        """
        # Different URI per framework
        uri = get_endpoint_uri(self.model.id, version_id)
        if self.framework == Framework.CAIRO:
            return f"{uri}/cairo_run"
        else:
            return f"{uri}/predict"

    def _get_model(self, model_id: int):
        """
        Retrieves the model specified by model_id.

        Args:
            model_id (int): The unique identifier of the model.

        Returns:
            The model.
        """
        return self.model_client.get(model_id)

    def _get_version(self, version_id: int):
        """
        Retrieves the version of the model specified by model id and version id.

        Args:
            version_id (int): The version number of the model.

        Returns:
            The version of the model.
        """
        return self.version_client.get(self.model.id, version_id)

    def _set_session(self):
        """
        Set onnxruntime session for the model specified by model id.

        Raises:
            ValueError: If the model version status is not completed.
        """

        if self.version.status != VersionStatus.COMPLETED:
            raise ValueError(
                f"Model version status is not completed {self.version.status}"
            )

        try:
            onnx_model = self.version_client.download_original(
                self.model.id, self.version.version)

            return ort.InferenceSession(onnx_model)

        except Exception as e:
            print(f"Could not download model: {e}")
            return None

    def _download_model(self, output_path: str):
        """
        Downloads the model specified by model id and version id to the given output_path.

        Args:
            output_path (str): The file path where the downloaded model should be saved.

        Raises:
            ValueError: If the model version status is not completed.
        """

        if self.version.status != VersionStatus.COMPLETED:
            raise ValueError(
                f"Model version status is not completed {self.version.status}"
            )

        onnx_model = self.version_client.download_original(
            self.model.id, self.version.version)

        print("ONNX model is ready, downloading! ✅")

        if ".onnx" in output_path:
            save_path = Path(output_path)
        else:
            save_path = Path(f"{output_path}/{self.model.name}.onnx")

        with open(save_path, "wb") as f:
            f.write(onnx_model)

        print(f"ONNX model saved at: {save_path} ✅")

    def _get_credentials(self):
        """
        Retrieves and sets the necessary credentials for API access.
        """
        self.api_client.retrieve_token()
        self.api_client.retrieve_api_key()

    def predict(
        self,
        input_file: Optional[str] = None,
        input_feed: Optional[Dict] = None,
        verifiable: bool = False,
        fp_impl="FP16x16",
        custom_output_dtype: Optional[str] = None,
        job_size: str = "M",
    ):
        """
        Makes a prediction using either a local ONNX session or a remote deployed model, depending on the
        instance configuration.

        Args:
            input_file (Optional[str]): The path to the input file for prediction. Defaults to None.
            input_feed (Optional[Dict]): A dictionary containing the input data for prediction. Defaults to None.
            verifiable (bool): A flag indicating whether to use the verifiable computation endpoint. Defaults to False.
            fp_impl (str): The fixed point implementation to use, when computed in verifiable mode. Defaults to "FP16x16".
            custom_output_dtype (Optional[str]): Specify the data type of the result when computed in verifiable mode. Defaults to None.

        Returns:
            A tuple (predictions, request_id) where predictions is the result of the prediction and request_id
            is the identifier of the prediction request if verifiable computation is used, otherwise None.

        Raises:
            ValueError: If required parameters are not provided or the session is not initialized.
        """
        try:
            if verifiable:
                if not self.uri:
                    raise ValueError("Model has not been deployed")

                # Non common arguments should be named parameters
                payload = self._format_inputs_for_framework(
                    input_file, input_feed, fp_impl=fp_impl, job_size=job_size
                )

                response = requests.post(self.uri, json=payload)

                try:
                    response.raise_for_status()
                except requests.exceptions.HTTPError as e:
                    logging.error(f"An error occurred in predict: {e}")
                    error_message = f"Deployment predict error: {response.text}"
                    logging.error(error_message)
                    raise e

                body = response.json()
                serialized_output = body["result"]
                request_id =  body["request_id"]

                if self.framework == Framework.CAIRO:
                    logging.info("Serialized: ", serialized_output)

                    if custom_output_dtype is None:
                        output_dtype = self._get_output_dtype()
                    else:
                        output_dtype = custom_output_dtype

                    preds = self._parse_cairo_response(
                        serialized_output, output_dtype)
                elif self.framework == Framework.EZKL:
                    preds = np.array(serialized_output[0])
                return (preds, request_id)

            else:
                if self.session is None:
                    raise ValueError("Session is not initialized.")
                if input_feed is None:
                    raise ValueError("Input feed is none")
                preds = self.session.run(None, input_feed)[0]
                return preds
        except Exception as e:
            logging.error(f"An error occurred in predict: {e}")
            raise e

    def _format_inputs_for_framework(self, *args, **kwargs):
        """
        Formats the inputs for a prediction request for a specific framework.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        match self.framework:
            case Framework.CAIRO:
                return self._format_inputs_for_cairo(*args, **kwargs)
            case Framework.EZKL:
                return self._format_inputs_for_ezkl(*args, **kwargs)
            case _:
                # This should never happen
                raise ValueError(f"Unsupported framework: {self.framework}")

    def _format_inputs_for_cairo(
        self,
        input_file: Optional[str],
        input_feed: Optional[Dict],
        fp_impl,
        job_size: str,
    ):
        """
        Formats the inputs for a prediction request for OrionRunner.

        Args:
            input_file (Optional[str]): The path to the input file for prediction. Defaults to None.
            input_feed (Optional[Dict]): A dictionary containing the input data for prediction. Defaults to None.
            fp_impl (str): The fixed point implementation to use.

        Returns:
            dict: A dictionary representing the formatted inputs for the Cairo prediction request.
        """
        serialized = []

        if input_file is not None:
            serialized = serialize(input_file, fp_impl)

        if input_feed is not None:
            for name in input_feed:
                value = input_feed[name]
                if isinstance(value, np.ndarray):
                    tensor = create_tensor_from_array(value, fp_impl)
                    serialized.append(serializer(tensor))
                else:
                    serialized.append(serializer(tensor))

        return {"job_size": job_size, "args": " ".join(serialized)}

    def _format_inputs_for_ezkl(
        self, input_file: str, input_feed: Dict, job_size: str, *args, **kwargs
    ):
        """
        Formats the inputs for a prediction request for EZKL.

        Args:
            input_file (str): The path to the input file for prediction.
            input_feed (Dict): A dictionary containing the input data for prediction.

        Returns:
            dict: A dictionary representing the formatted inputs for the EZKL prediction request.
        """
        if input_file is not None:
            data = load_data(input_file).reshape([-1])
        elif input_feed is not None:
            match input_feed:
                case dict():
                    data = input_feed["input_data"]
                case list():
                    data = input_feed
                case np.ndarray():
                    data = input_feed.reshape([-1])
                case _:
                    raise ValueError(
                        "Invalid input_feed format. Must be a dictionary with 'input_data' containintg the data array."
                    )
        return {"input_data": [data], "job_size": job_size}

    def _parse_cairo_response(self, response, data_type: str):
        """
        Parses the response from a OrionRunner prediction request.

        Args:
            response (str): The serialized response from the Cairo prediction request.
            data_type (str): The data type to which the response should be deserialized.
            fp_impl (str): The fixed point implementation used.

        Returns:
            The deserialized prediction result.
        """
        return deserialize(response, data_type)

    def _get_output_dtype(self):
        """
        Retrieve the Cairo output data type base on the operator type of the final node.

        Returns:
            The output dtype as a string.
        """

        file = self.version_client.download_original(
            self.model.id, self.version.version
        )

        model = onnx.load_model_from_string(file)
        graph = model.graph
        output_tensor_name = graph.output[0].name

        def find_producing_node(graph, tensor_name):
            for node in graph.node:
                if tensor_name in node.output:
                    return node
            return None

        final_node = find_producing_node(graph, output_tensor_name)
        optype = final_node.op_type

        match optype:
            case "TreeEnsembleClassifier":
                return "(Span<u32>, MutMatrix<FP16x16>)"

            case "TreeEnsembleRegressor":
                return "MutMatrix::<FP16x16>"

            case "LinearClassifier":
                return "(Span<u32>, Tensor<FP16x16>)"
            case _:
                return "Tensor<FP16x16>"



import logging


class WordReplacerFilter(logging.Filter):
    def filter(self, record):
        record.msg = record.msg.replace("flow", "action")
        record.msg = record.msg.replace("Flow", "Action")
        return True



from functools import partial, wraps

from prefect import task as prefect_task


def task(func=None, *task_init_args, **task_init_kwargs):
    if func is None:
        return partial(task, *task_init_args, **task_init_kwargs)

    @wraps(func)
    def safe_func(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            return res
        except Exception as e:
            raise e

    safe_func.__name__ = func.__name__
    return prefect_task(safe_func, *task_init_args, **task_init_kwargs)



import os  # noqa: E402

from giza_actions.utils import get_workspace_uri  # noqa: E402

os.environ["PREFECT_API_URL"] = f"{get_workspace_uri()}/api"
os.environ["PREFECT_UI_URL"] = get_workspace_uri()

from prefect.deployments import run_deployment  # noqa: E402


def run_action_deployment(name: str, parameters: dict = None):
    deployment_run = run_deployment(name=name, parameters=parameters)
    print(
        f"Deployment run name: {deployment_run.name} exited with state: {deployment_run.state_name}"
    )
    return deployment_run



import pathlib

# The absolute path to this module
__module_path__ = pathlib.Path(__file__).parent
# The absolute path to the root of the repository, only valid for use during development
__development_base_path__ = __module_path__.parents[1]
